@@ -1,15 +1,20 @@
 """Asynchronous Session Module (:mod:`pymanip.asyncsession`)
 ============================================================
 
-This module defines two classes, :class:`pymanip.asyncsession.AsyncSession`
-and :class:`pymanip.asyncsession.RemoteObserver`. The former is used to
+This module defines two classes for live acquisition, :class:`~pymanip.asyncsession.AsyncSession`
+and :class:`~pymanip.asyncsession.RemoteObserver`. The former is used to
 manage an experimental session, the latter to access its live data from a
-remote computer.
+remote computer. There is also one class for read-only access to previous session,
+:class:`~pymanip.asyncsession.SavedAsyncSession`.
 
 .. autoclass:: AsyncSession
    :members:
    :private-members:
 
+.. autoclass:: SavedAsyncSession
+   :members:
+   :private-members:
+
 .. autoclass:: RemoteObserver
    :members:
    :private-members:
@@ -26,6 +31,12 @@ import pickle
 import warnings
 import inspect
 from pprint import pprint
+from functools import lru_cache
+
+try:
+    from functools import cached_property
+except ImportError:
+    from backports.cached_property import cached_property
 
 import sqlite3
 from datetime import datetime
@@ -45,7 +56,7 @@ import json
 
 try:
     import PyQt5.QtCore
-except ModuleNotFoundError:
+except (ModuleNotFoundError, FileNotFoundError):
     pass
 
 from fluiddyn.util.terminal_colors import cprint
@@ -73,35 +84,92 @@ class AsyncSession:
 
     database_version = 3
 
-    def __init__(self, session_name=None, verbose=True, delay_save=False):
-        """Constructor method
-        """
+    def __init__(
+        self,
+        session_name=None,
+        verbose=True,
+        delay_save=False,
+        exist_ok=True,
+        readonly=False,
+    ):
+        """Constructor method"""
 
-        self.session_name = session_name
-        self.custom_figures = None
-        self.delay_save = delay_save
         if session_name is not None:
             session_name = str(session_name)  # in case it is a Path object
             if session_name.endswith(".db"):
                 session_name = session_name[:-3]
         elif delay_save:
             raise ValueError("Cannot delay_save if session_name is not specified")
-        if session_name is None or delay_save:
+
+        self.session_name = session_name
+        self.verbose = verbose
+        self.exist_ok = exist_ok
+        self.readonly = readonly
+        self.delay_save = delay_save
+
+        self.custom_figures = None
+        self.figure_list = []
+        self.template_dir = os.path.join(os.path.dirname(__file__), "web")
+        self.static_dir = os.path.join(os.path.dirname(__file__), "web_static")
+        self.jinja2_loader = jinja2.FileSystemLoader(self.template_dir)
+        self.conn = None
+
+    def save_database(self):
+        """This method is useful only if delay_save = True. Then, the database is kept in-memory for
+        the duration of the session. This method saves the database on the disk.
+        A new database file will be created with the content of the current
+        in-memory database.
+
+        This method is automatically called at the exit of the context manager.
+        """
+        if self.conn is None:
+            raise RuntimeError("AsyncSession is not opened!")
+
+        if self.delay_save:
+            try:
+                os.remove(str(self.session_name) + ".db")
+            except FileNotFoundError:
+                pass
+            disk_db = sqlite3.connect(str(self.session_name) + ".db")
+            try:
+                with disk_db as c:
+                    for line in self.conn.iterdump():
+                        c.execute(line)
+            finally:
+                disk_db.close()
+
+    def open(self):
+        """Opens database for reading or writting"""
+
+        if self.conn is not None:
+            raise RuntimeError("Already opened!")
+
+        if not self.exist_ok and os.path.exists(self.session_name + ".db"):
+            raise RuntimeError("File exists !")
+
+        if self.session_name is None or self.delay_save:
             # For no name session, or in case of delay_save=True, then
             # the connection is in-memory
             self.conn = sqlite3.connect(":memory:")
         else:
-            # Otherwise, the connection is on the disk for immediate writing
-            self.conn = sqlite3.connect(session_name + ".db")
-        if delay_save and os.path.exists(session_name + ".db"):
+            # Otherwise, the connection is on the disk for immediate read/write
+            if self.readonly:
+                uri = "file:{path:}?mode=ro".format(path=self.session_name + ".db")
+                self.conn = sqlite3.connect(uri, uri=True)
+            else:
+                self.conn = sqlite3.connect(self.session_name + ".db")
+
+        if self.delay_save and os.path.exists(self.session_name + ".db"):
             # Load existing database into in-memory database
-            disk_db = sqlite3.connect(session_name + ".db")
+            uri = "file:{path:}?mode=ro".format(path=self.session_name + ".db")
+            disk_db = sqlite3.connect(uri, uri=True)
             try:
                 with self.conn as c:
                     for line in disk_db.iterdump():
                         c.execute(line)
             finally:
                 disk_db.close()
+
         with self.conn as c:
             tables = list(c.execute("SELECT name FROM sqlite_master;"))
             if not tables:
@@ -156,81 +224,53 @@ class AsyncSession:
                     """,
                     ("_session_creation_timestamp", datetime.now().timestamp()),
                 )
-            elif verbose:
+            elif self.verbose:
                 self.print_welcome()
-        self.figure_list = []
-        self.template_dir = os.path.join(os.path.dirname(__file__), "web")
-        self.static_dir = os.path.join(os.path.dirname(__file__), "web_static")
-        self.jinja2_loader = jinja2.FileSystemLoader(self.template_dir)
 
-    def save_database(self):
-        """This method is useful only if delay_save = True. Then, the database is kept in-memory for
-        the duration of the session. This method saves the database on the disk.
-        A new database file will be created with the content of the current
-        in-memory database.
-
-        This method is automatically called at the exit of the context manager.
-        """
-        if self.delay_save:
-            try:
-                os.remove(self.session_name + ".db")
-            except FileNotFoundError:
-                pass
-            disk_db = sqlite3.connect(self.session_name + ".db")
-            try:
-                with disk_db as c:
-                    for line in self.conn.iterdump():
-                        c.execute(line)
-            finally:
-                disk_db.close()
+    def close(self):
+        if self.conn:
+            self.save_database()
+            self.conn.close()
+        self.conn = None
 
     def __enter__(self):
-        """Context manager enter method
-        """
+        """Context manager enter method"""
+        if not self.conn:
+            self.open()
         return self
 
     def __exit__(self, type_, value, cb):
-        """Context manager exit method
-        """
-        self.save_database()
-        self.conn.close()
+        """Context manager exit method"""
+        self.close()
 
     def get_version(self):
-        """Returns current version of the database layout.
-        """
+        """Returns current version of the database layout."""
         version = self.parameter("_database_version")
         if version is None:
             version = 1
         return version
 
-    @property
+    @cached_property
     def t0(self):
-        """Session creation timestamp
-        """
-        if hasattr(self, "_session_creation_timestamp"):
-            return self._session_creation_timestamp
+        """Session creation timestamp"""
         t0 = self.parameter("_session_creation_timestamp")
         if t0 is not None:
-            self._session_creation_timestamp = t0
             return t0
         logged_data = self.logged_first_values()
         if logged_data:
             t0 = min([v[0] for k, v in logged_data.items()])
             self.save_parameter(_session_creation_timestamp=t0)
-            self._session_creation_timestamp = t0
             return t0
         return 0
 
     @property
     def initial_timestamp(self):
-        """Session creation timestamp, identical to :attr:`pymanip.asyncsession.AsyncSession.t0`
-        """
+        """Session creation timestamp, identical to :attr:`pymanip.asyncsession.AsyncSession.t0`"""
         return self.t0
 
     @property
     def last_timestamp(self):
-        """Timestamp of the last recorded value
-        """
+        """Timestamp of the last recorded value"""
         ts = list()
         last_values = self.logged_last_values()
         if last_values:
@@ -253,8 +293,7 @@ class AsyncSession:
             cprint.blue("***   End date: " + end_string)
 
     def print_description(self):
-        """Prints the list of parameters, logged variables and datasets.
-        """
+        """Prints the list of parameters, logged variables and datasets."""
         version = self.get_version()
         print(
             self.session_name,
@@ -545,7 +584,8 @@ class AsyncSession:
         :return: dataset value
         :rtype: object
         """
-        return next(self.datasets(name))
+        last_ts = self.dataset_times(name)[-1]
+        return last_ts, self.dataset(name, ts=last_ts)
 
     def dataset_times(self, name):
         """This method returns the timestamp of the recorded dataset under the specified
@@ -703,10 +743,14 @@ class AsyncSession:
         from_addr,
         to_addrs,
         host,
-        port=25,
+        port=None,
         subject=None,
         delay_hours=6,
         initial_delay_hours=None,
+        use_ssl_submission=False,
+        use_starttls=False,
+        user=None,
+        password=None,
     ):
         """This method returns an asynchronous task which sends an email at regular intervals.
         Such a task should be passed to :meth:`pymanip.asyncsession.AsyncSession.monitor` or
@@ -726,6 +770,15 @@ class AsyncSession:
         :type initial_delay_hours: float, optional
         """
 
+        if port is None:
+            if use_ssl_submission:
+                port = 465
+            elif use_starttls:
+                # Tough some servers can do starttls on port 25, but then the user must specify port=25
+                # as this is less common than on the submission port.
+                port = 587
+            else:
+                port = 25
         if self.session_name is None:
             title = "Pymanip session"
         else:
@@ -796,7 +849,35 @@ class AsyncSession:
                     filename="fig{:d}-{:}.png".format(fignum, datestr),
                 )
 
-            with smtplib.SMTP(host, port) as smtp:
+            try:
+                if use_ssl_submission:
+                    smtp_server = smtplib.SMTP_SSL(host, port)
+                else:
+                    smtp_server = smtplib.SMTP(host, port)
+            except Exception as e:
+                print("Unable to connect to SMTP server")
+                print(e)
+                await self.sleep(60, verbose=False)
+                continue
+
+            with smtp_server as smtp:
+                if use_starttls:
+                    smtp.starttls()
+                if user and password:
+                    if not use_starttls and not use_ssl_submission:
+                        raise RuntimeError(
+                            "Do you really want to send password unencrypted?"
+                        )
+                    try:
+                        smtp.login(user, password)
+                    except smtplib.SMTPHeloError:
+                        print("The server didn’t reply properly to the HELO greeting.")
+                    except smtplib.SMTPAuthenticationError:
+                        print(
+                            "The server didn’t accept the username/password combination."
+                        )
+                    except smtplib.SMTPNotSupportedError:
+                        print("The AUTH command is not supported by the server.")
                 try:
                     smtp.send_message(msg)
                     print("Email sent!")
@@ -866,11 +947,13 @@ class AsyncSession:
         saved_figsize = self.parameter(param_key_figsize)
         if saved_figsize:
             saved_figsize = eval(saved_figsize)
-        plt.ion()
+        if not self.offscreen_figures:
+            plt.ion()
         fig = plt.figure(figsize=saved_figsize)
-        mngr = fig.canvas.manager
-        if saved_geom:
-            mngr.window.setGeometry(saved_geom)
+        if not self.offscreen_figures:
+            mngr = fig.canvas.manager
+            if saved_geom:
+                mngr.window.setGeometry(saved_geom)
         ax = fig.add_subplot(111)
         line_objects = dict()
         self.figure_list.append(fig)
@@ -928,7 +1011,8 @@ class AsyncSession:
                                 pass
                         else:
                             ax.set_ylim(fixed_ylim)
-                        fig.show()
+                        if not self.offscreen_figures:
+                            fig.show()
                     last_update[x] = ts_x[-1]
                     last_update[y] = ts_y[-1]
             else:
@@ -972,38 +1056,43 @@ class AsyncSession:
                             if fixed_ylim is not None:
                                 ax.set_ylim(fixed_ylim)
                             ax.legend()
-                            fig.show()
+                            if not self.offscreen_figures:
+                                fig.show()
                         last_update[name] = ts[-1]
             await asyncio.sleep(1)
 
-        # Saving figure positions
-        try:
-            geom = mngr.window.geometry()
-            figsize = tuple(fig.get_size_inches())
-            self.save_parameter(
-                **{param_key_window: str(geom), param_key_figsize: str(figsize)}
-            )
-        except AttributeError:
-            pass
+        if not self.offscreen_figures:
+            # Saving figure positions
+            try:
+                geom = mngr.window.geometry()
+                figsize = tuple(fig.get_size_inches())
+                self.save_parameter(
+                    **{param_key_window: str(geom), param_key_figsize: str(figsize)}
+                )
+            except AttributeError:
+                pass
 
     async def figure_gui_update(self):
         """This method returns an asynchronous task which updates the figures created by the
         :meth:`pymanip.asyncsession.AsyncSession.plot` tasks. This task is added automatically,
         and should not be used manually.
         """
-        with warnings.catch_warnings():
-            warnings.filterwarnings("ignore", category=MatplotlibDeprecationWarning)
-            while self.running:
-                figure_list = self.figure_list
-                if self.custom_figures:
-                    figure_list = figure_list + self.custom_figures
-                if figure_list:
-                    for fig in self.figure_list:
-                        fig.canvas.start_event_loop(0.7 / len(self.figure_list))
-                        await asyncio.sleep(0.3 / len(self.figure_list))
-                    await asyncio.sleep(0.05)
-                else:
-                    await asyncio.sleep(1.0)
+        if not self.offscreen_figures:
+            with warnings.catch_warnings():
+                warnings.filterwarnings("ignore", category=MatplotlibDeprecationWarning)
+                while self.running:
+                    figure_list = self.figure_list
+                    if self.custom_figures:
+                        figure_list = figure_list + self.custom_figures
+                    if figure_list:
+                        for fig in self.figure_list:
+                            fig.canvas.start_event_loop(0.7 / len(self.figure_list))
+                            await asyncio.sleep(0.3 / len(self.figure_list))
+                        await asyncio.sleep(0.05)
+                    else:
+                        await asyncio.sleep(1.0)
+        else:
+            await asyncio.sleep(0.0)
 
     def ask_exit(self, *args, **kwargs):
         """This methods informs all tasks that the monitoring session should stop. Call this method if you
@@ -1099,7 +1188,7 @@ class AsyncSession:
         data = [
             {
                 "name": name,
-                "value": v[1],
+                "value": str(v[1]) if isinstance(v[1], bytes) else v[1],
                 "datestr": time.strftime(dateformat, time.localtime(v[0])),
             }
             for name, v in self.logged_last_values().items()
@@ -1110,7 +1199,11 @@ class AsyncSession:
         """This asynchronous method returns the HTTP response to a request for JSON data of the session
         parameters. Should not be called manually.
         """
-        params = {k: v for k, v in self.parameters().items() if not k.startswith("_")}
+        params = {
+            k: (str(v) if isinstance(v, bytes) else v)
+            for k, v in self.parameters().items()
+            if not k.startswith("_")
+        }
         return web.json_response(params)
 
     async def server_plot_page(self, request):
@@ -1155,7 +1248,12 @@ class AsyncSession:
         print("Task finished", corofunc)
 
     async def monitor(
-        self, *tasks, server_port=6913, custom_routes=None, custom_figures=None
+        self,
+        *tasks,
+        server_port=6913,
+        custom_routes=None,
+        custom_figures=None,
+        offscreen_figures=False,
     ):
         """This method runs the specified tasks, and opens a web-server for remote access and set up the tasks to run
         matplotlib event loops if necessary. This is the main
@@ -1174,9 +1272,12 @@ class AsyncSession:
         :type custom_routes: co-routine function, optional
         :param custom_figures: additional matplotlib figure object that needs to run the matplotlib event loop
         :type custom_figures: :class:`matplotlib.pyplot.Figure`, optional
+        :param offscreen_figures: if set, figures are not shown onscreen
+        :type offscreen_figures: bool, optional
         """
         loop = asyncio.get_event_loop()
         self.custom_figures = custom_figures
+        self.offscreen_figures = offscreen_figures
 
         # signal handling
         self.running = True
@@ -1225,9 +1326,15 @@ class AsyncSession:
         else:
             await asyncio.gather(self.figure_gui_update(), *tasks_final)
 
-    def run(self, *tasks, server_port=6913, custom_routes=None, custom_figures=None):
-        """Synchronous call to :meth:`pymanip.asyncsession.AsyncSession.monitor`.
-        """
+    def run(
+        self,
+        *tasks,
+        server_port=6913,
+        custom_routes=None,
+        custom_figures=None,
+        offscreen_figures=False,
+    ):
+        """Synchronous call to :meth:`pymanip.asyncsession.AsyncSession.monitor`."""
 
         asyncio.run(
             self.monitor(
@@ -1235,6 +1342,7 @@ class AsyncSession:
                 server_port=server_port,
                 custom_routes=custom_routes,
                 custom_figures=custom_figures,
+                offscreen_figures=offscreen_figures,
             )
         )
 
@@ -1283,14 +1391,12 @@ class RemoteObserver:
     """
 
     def __init__(self, host, port=6913):
-        """Constructor method
-        """
+        """Constructor method"""
         self.host = host
         self.port = port
 
     def _get_request(self, apiname):
-        """Private method to send a GET request for the specified API name
-        """
+        """Private method to send a GET request for the specified API name"""
         url = "http://{host:}:{port:}/api/{api:}".format(
             host=self.host, port=self.port, api=apiname
         )
@@ -1302,8 +1408,7 @@ class RemoteObserver:
             raise
 
     def _post_request(self, apiname, params):
-        """Private method to send a POST request for the specified API name and params
-        """
+        """Private method to send a POST request for the specified API name and params"""
         url = "http://{host:}:{port:}/api/{api:}".format(
             host=self.host, port=self.port, api=apiname
         )
@@ -1378,6 +1483,194 @@ class RemoteObserver:
         return recordings
 
 
+class SavedAsyncSession:
+    """This class implements the same methods as AsyncSession with readonly mode, but with
+    caching enabled. Also, the file is opened on demand. No context manager is necessary.
+    """
+
+    def __init__(self, session_name, verbose=True):
+
+        self.session_name = session_name
+        self.verbose = verbose
+        self.session = AsyncSession(session_name, verbose=False, readonly=True)
+        if verbose:
+            try:
+                self.print_welcome()
+            except sqlite3.OperationalError:
+                print("Warning: unable to open database for session", session_name)
+
+    def print_welcome(self):
+        """Prints informative start date/end date message. If verbose is True, this method
+        is called by the constructor.
+        """
+        start_string = time.strftime(dateformat, time.localtime(self.initial_timestamp))
+        cprint.blue("*** Start date: " + start_string)
+        last = self.last_timestamp
+        if last:
+            end_string = time.strftime(dateformat, time.localtime(last))
+            cprint.blue("***   End date: " + end_string)
+
+    def print_description(self):
+        """Prints the list of parameters, logged variables and datasets."""
+        version = self.get_version()
+        print(
+            self.session_name,
+            "is an asynchroneous session (version {:}).".format(version),
+        )
+        print()
+        last_values = self.logged_last_values()
+        params = {
+            key: val
+            for key, val in self.parameters().items()
+            if not key.startswith("_")
+        }
+        if params:
+            print("Parameters")
+            print("==========")
+            for key, val in self.parameters().items():
+                print(key, ":", val)
+            print()
+
+        if last_values:
+            print("Logged variables")
+            print("================")
+            for name, t_v in last_values.items():
+                print(name, "(", t_v[1], ")")
+            print()
+
+        ds_names = self.dataset_names()
+        if ds_names:
+            print("Datasets")
+            print("========")
+            for ds in ds_names:
+                print(ds)
+            print()
+
+    # General attributes
+
+    @lru_cache(maxsize=128)
+    def get_version(self):
+        with self.session as sesn:
+            return sesn.get_version()
+
+    @property
+    def t0(self):
+        with self.session as sesn:
+            return sesn.t0
+
+    @property
+    def initial_timestamp(self):
+        return self.t0
+
+    @cached_property
+    def last_timestamp(self):
+        ts = list()
+        last_values = self.logged_last_values()
+        if last_values:
+            ts.append(max([t_v[0] for name, t_v in last_values.items()]))
+        for ds_name in self.dataset_names():
+            ts.append(max(self.dataset_times(ds_name)))
+        if ts:
+            return max(ts)
+        return None
+
+    # Logged variables
+
+    @lru_cache(maxsize=128)
+    def logged_data(self):
+        with self.session as sesn:
+            return sesn.logged_data()
+
+    def logged_variables(self):
+        return set(self.logged_data().keys())
+
+    def logged_variable(self, varname):
+        return self.logged_data()[varname]
+
+    @lru_cache(maxsize=128)
+    def logged_last_values(self):
+        last = dict()
+        for name, data in self.logged_data().items():
+            ts, val = data
+            max_ts = np.max(ts)
+            assert ts[-1] == max_ts
+            last[name] = (ts[-1], val[-1])
+        return last
+
+    @lru_cache(maxsize=128)
+    def logged_first_values(self):
+        first = dict()
+        for name, data in self.logged_data().items():
+            ts, val = data
+            min_ts = np.min(ts)
+            assert ts[0] == min_ts
+            first[name] = (ts[0], val[0])
+        return first
+
+    @lru_cache(maxsize=128)
+    def logged_data_fromtimestamp(self, name, timestamp):
+        ts, val = self.logged_variable(name)
+        for ind, tt in enumerate(ts):
+            if tt == timestamp:
+                break
+        else:
+            raise ValueError("No logged data at specified timestamp")
+        return val[ind]
+
+    def __getitem__(self, key):
+        return self.logged_variable(key)
+
+    # Dataset
+
+    @lru_cache(maxsize=128)
+    def dataset_names(self):
+        with self.session as sesn:
+            return sesn.dataset_names()
+
+    def datasets(self, name):
+        with self.session as sesn:
+            return sesn.datasets()
+
+    def dataset_last_data(self, name):
+        """This method returns the last recorded dataset under the specified name.
+
+        :param name: name of the dataset to retrieve
+        :type name: str
+        :return: dataset value
+        :rtype: object
+        """
+        last_ts = self.dataset_times(name)[-1]
+        return last_ts, self.dataset(name, ts=last_ts)
+
+    @lru_cache(maxsize=128)
+    def dataset_times(self, name):
+        with self.session as sesn:
+            sesn.dataset_times(name)
+
+    def dataset(self, name, ts=None, n=None):
+        if ts is None:
+            if n is None:
+                ts, data = self.dataset_last_data(name)
+                return data
+            else:
+                ts = self.dataset_times(name)[n]
+        with self.session as sesn:
+            return sesn.dataset(name, ts=ts, n=n)
+
+    # Parameters
+
+    @lru_cache(maxsize=128)
+    def parameters(self):
+        with self.session as sesn:
+            return sesn.parameters()
+
+    def parameter(self, name):
+        return self.parameters()[name]
+
+    def has_parameter(self, name):
+        return name in self.parameters()
+
+
 if __name__ == "__main__":
     with AsyncSession("Essai") as sesn:
         sesn.add_entry(a=1, b=2)
